//#ifdef __GNUG__
//#pragma implementation
//#endif
//#include "wx/wxprec.h"

//#ifndef  WX_PRECOMP
  #include "wx/wx.h"
  #include <wx/listctrl.h>
//#endif //precompiled headers


#include "main.h"
#include "main_frame.h"
#include "algorithm.h"
#include "alg_try.h"
#include "video.h"
#include "image_scope.h"
#include "image_object.h"

IMPLEMENT_APP(PixalcoApp)

const int DEFAULT_IMAGE_BUFFER_SIZE = 5;
const int DEFAULT_MOVIE_BUFFER_SIZE = 10;

PixalcoApp::PixalcoApp()
{
    mImageBufferSize = DEFAULT_IMAGE_BUFFER_SIZE;
    mMovieBufferSize = DEFAULT_MOVIE_BUFFER_SIZE;
    mMainFrame = NULL;
    VideoManager::Create();
    
    // TODO: needs to support each line and row
    // when that is supported, this has to be moved
    for ( int i=0; i<MAX_ROWS; ++i )
    {
        mImageObjBuffer[i] = new list<ImageObject*>();
    }
}

PixalcoApp::~PixalcoApp()
{
    ClearMovieBuffer();
    for ( int i=0; i<MAX_ROWS; ++i )
    {
        ClearBuffer(0, i);
        delete mImageObjBuffer[i];
    }
       
    RemoveAllAlgorithms();
    RemoveAllMediaFiles();
    VideoManager::Destroy();
}

bool PixalcoApp::OnInit()
{
    ::wxInitAllImageHandlers();
    
    mMainFrame = new MainFrame("Pixalco", 50, 50, 450, 300);

    Algorithm * alg = new AlgTry1();
    AddAlgorithm(alg);
    alg = new AlgTry2();
    AddAlgorithm(alg);
    alg = new AlgTry3();
    AddAlgorithm(alg);
    alg = new RGBtoBlackWhiteAvg();
    AddAlgorithm(alg);
    alg = new RGBtoBlackWhiteNTSC();
    AddAlgorithm(alg);
    alg = new MovieDiffImage();
    AddAlgorithm(alg);

    mMainFrame->Show(TRUE);
    SetTopWindow(mMainFrame);
    return TRUE;	
}

void PixalcoApp::AddMediaFile(const string& displayName_, const string& fileName_)
{
    wxListCtrl &listImages = mMainFrame->AccessImagesList();
        
    string name(displayName_);
    long id=listImages.FindItem(-1, name.c_str());
    while ( id >= 0 )
    {
        name += "_";
        id = listImages.FindItem(-1, name.c_str());
    }

    ImageObject * imageObj = new ImageObject(name);
    
    if (    imageObj->GetImage()->CanRead( fileName_.c_str() )
            && imageObj->GetImage()->LoadFile( fileName_.c_str() ) )
    {
        MediaFile media;
        media.mMediaType = MEDIA_IMAGE;
        media.mFileName = fileName_;
        media.mDisplayName = displayName_;
        media.mImageObject = imageObj;
        mMediaFiles.push_back(media);
        listImages.InsertItem( listImages.GetItemCount(), name.c_str() );
    }
    else // video support
    {
        Video * video = new Video();
        if ( video->LoadFile(fileName_.c_str()) )
        {
            int width, height;
            video->GetSize(width, height);
            imageObj->GetImage()->Create(width, height);
            MediaFile media;
            media.mMediaType = MEDIA_MOVIE;
            media.mFileName = fileName_;
            media.mDisplayName = displayName_;
            media.mVideo = video;
            media.mImageObject = imageObj;
            mMediaFiles.push_back(media);
            listImages.InsertItem( listImages.GetItemCount(), name.c_str() );
        }
        else
        {
            imageObj->Release();
            delete video;
        }
    }
}

MediaFile * PixalcoApp::GetMediaFile(const string& displayName_)
{
    for (vector<MediaFile>::iterator it = mMediaFiles.begin(); it != mMediaFiles.end(); it++)
    {
        if ( it->mDisplayName == displayName_ )
        {
            return &(*it);
        }
    }
    return NULL;
}

void PixalcoApp::RemoveMediaFile(const string& displayName_)
{
    for (vector<MediaFile>::iterator it = mMediaFiles.begin(); it != mMediaFiles.end(); it++)
    {
        if ( it->mDisplayName == displayName_ )
        {
            wxListCtrl &listImages = mMainFrame->AccessImagesList();
            long id=listImages.FindItem(-1, displayName_.c_str());            
            if ( id >= 0 )
                listImages.DeleteItem(id);
            
            delete it->mVideo;
            if ( it->mImageObject )
                it->mImageObject->Release();
            mMediaFiles.erase(it);
            
            return;
        }
    }
}

void PixalcoApp::SelectMediaFile(const string& displayName_)
{
    MediaFile * info = GetMediaFile(displayName_);
    SetMediaFile(info);
}

// TODO: think about movie stepping
void PixalcoApp::SetMediaFile(MediaFile * media_)
{
    if ( !media_ )
        return;
    
    ImageObject * old = mMainFrame->GetImageScope(0,0)->GetImageObject();        
    if ( media_->GetMediaType() ==  MEDIA_MOVIE )
    {

        media_->mVideo->GetNextFrame( media_->mImageObject->GetImage() );
        
        ImageObject * copy = media_->mImageObject->Copy();
        AddImageObjToMovieBuffer(copy);
        copy->Release();

        // TODO
//       mMainFrame->UpdateMovieScrollerPos(media_->mVideo);
    }
    else if ( old && old != media_->mImageObject )
    {
        AddImageObjToBuffer(old, 0, 0);
    }

    mMainFrame->GetImageScope(0,0)->SetImageObject(media_->mImageObject);
}

void PixalcoApp::RemoveAllMediaFiles(bool updateMainFrame_)
{
    wxListCtrl &listImages = mMainFrame->AccessImagesList();
    for (vector<MediaFile>::iterator it = mMediaFiles.begin(); it != mMediaFiles.end(); it++)
    {
        if ( updateMainFrame_ )
        {
            long id=listImages.FindItem(-1, it->mDisplayName.c_str());
            if ( id >= 0 )
                listImages.DeleteItem(id);
        }
            
         delete it->mVideo;
         if ( it->mImageObject )
            it->mImageObject->Release();
    }
    mMediaFiles.clear();
}

void PixalcoApp::RunMovie(const string& displayName_, const string& algorithmName_)
{
    // TODO
}

void PixalcoApp::StopMovie(const string& displayName_)
{
    // TODO
}

// TODO: maybe - think about movie stepping
void PixalcoApp::StepMovie(const string& displayName_, const string& algorithmName_)
{
    MediaFile * media = GetMediaFile(displayName_);
    if ( !media )
        return;

    SetMediaFile(media);
    PIXALCO.ExecuteAlgorithm(algorithmName_);
}

void PixalcoApp::SeekMovie(const string& displayName_, long pos_)
{
    MediaFile * media = PIXALCO.GetMediaFile(displayName_);

    if ( media && media->GetMediaType() == MEDIA_MOVIE )
    {
        media->mVideo->Seek(pos_);
        media->mVideo->GetNextFrame( media->mImageObject->GetImage() );
        mMainFrame->UpdateMovieScrollerPos(media->mVideo);
        mMainFrame->GetImageScope(0,0)->SetImageObject(media->mImageObject);
    }
}

void PixalcoApp::AddAlgorithm(Algorithm *algorithm_)
{
    if ( !algorithm_ )
        return;
    
    wxListCtrl &listAlgorithms = mMainFrame->AccessAlgorithmsList();
    listAlgorithms.InsertItem( 0, algorithm_->GetName().c_str() );
    
    mAlgorithms.push_back(algorithm_);
}

Algorithm* PixalcoApp::GetAlgorithm(const string& name_)
{
    for(vector<Algorithm*>::iterator it = mAlgorithms.begin(); it != mAlgorithms.end(); it++ )
    {
        if ( (*it)->GetName() == name_ )
        {
            return (*it);
        }
    }
    return NULL;
}

void PixalcoApp::RemoveAlgorithm(const string& name_)
{
    for(vector<Algorithm*>::iterator it = mAlgorithms.begin(); it != mAlgorithms.end(); it++ )
    {
        if ( (*it)->GetName() == name_ )
        {
            wxListCtrl &listAlgorithms = mMainFrame->AccessAlgorithmsList();
            long id=listAlgorithms.FindItem(-1, name_.c_str());            
            if ( id >= 0 )
                listAlgorithms.DeleteItem(id);
                        
            delete (*it);
            mAlgorithms.erase(it);
            return;
        }
    }
}

void PixalcoApp::RemoveAllAlgorithms()
{
    for(vector<Algorithm*>::iterator it = mAlgorithms.begin(); it != mAlgorithms.end(); it++ )
        delete *it;
    mAlgorithms.clear();
}

void PixalcoApp::ExecuteAlgorithm(const string& name_)
{
    Algorithm * alg = GetAlgorithm( name_.c_str() );
    if ( !alg )
    {
        if ( name_.length() )
            fprintf(stderr,"PixalcoApp::ExecuteAlgorithm: algorithm not found");
        return;
    }

    int line = 0;

    ImageScope * scopes[3];
    ImageObject * copies[3];
    for ( int i=0; i<3; i++ )
    {
        copies[i] = NULL;
        scopes[i] = mMainFrame->GetImageScope(line,i);
        if ( !scopes[i] )
        {
            fprintf(stderr,"PixalcoApp::ExecuteAlgorithm: empty scope %d", i);
            return;
        }
    }

    wxImage * image = scopes[0]->GetImage();
    if ( !image || !image->Ok() )
        return;

    // TODO: hack for prototype
    copies[0] = scopes[0]->GetImageObject()->Copy();

    if ( alg->DoesNeedNewImage( image, scopes[1]->GetImage(), 1 ) )
    {
        if ( !scopes[1]->GetImageObject() )
            scopes[1]->SetImageObject( new ImageObject(scopes[0]->GetImageObject()->GetDisplayName()) );
        else
            copies[1] = scopes[1]->GetImageObject()->Copy();
        scopes[1]->GetImageObject()->SetImage( alg->MakeNewImage(image, 1) );
    }
    if ( alg->DoesNeedNewImage( image, scopes[2]->GetImage(), 2 ) )
    {
        if ( !scopes[2]->GetImageObject() )
            scopes[2]->SetImageObject( new ImageObject(scopes[0]->GetImageObject()->GetDisplayName()) );
        else
            copies[2] = scopes[2]->GetImageObject()->Copy();
        scopes[2]->GetImageObject()->SetImage( alg->MakeNewImage(image, 2) );
    }

    switch ( alg->GetNrParams() )
    {
        case 1:
            alg->Execute(image);
            scopes[0]->Refresh();
            break;
        case 2:
            alg->Execute(image, scopes[1]->GetImage());
            scopes[0]->Refresh();
            scopes[1]->Refresh();
            break;
        case 3:
            alg->Execute(image, scopes[1]->GetImage(), scopes[2]->GetImage());
            scopes[0]->Refresh();
            scopes[1]->Refresh();
            scopes[2]->Refresh();
            break;
    }
    
    for ( int i=0; i<3; i++ )
    {
        if ( copies[i] )
        {
            AddImageObjToBuffer(copies[i], 0, i);
            copies[i]->Release();
            copies[i] = NULL;
        }
    }
}

void PixalcoApp::AddImageObjToBuffer(ImageObject * imgObj_, int line_, int row_)
{
    if ( !imgObj_ || row_ < 0 || row_ >= MAX_ROWS || line_ < 0 || line_ > 0 )
        return;

//    cerr << imgObj_->GetDisplayName();

    imgObj_->AddRef();
    mImageObjBuffer[row_]->push_front(imgObj_);
    if ( mImageObjBuffer[row_]->size() > mImageBufferSize )
    {
        mImageObjBuffer[row_]->back()->Release();
        mImageObjBuffer[row_]->pop_back();
    } 
}

ImageObject* PixalcoApp::GetImageObjFromBuffer(int line_, int row_, int age_)
{
    if ( row_ < 0 || row_ >= MAX_ROWS || line_ < 0 || line_ > 0 )
        return NULL;

    if ( mImageObjBuffer[row_]->size() <= age_ )
        return NULL;

    list<ImageObject*>::iterator it=mImageObjBuffer[row_]->begin();
    for ( int i=0; i < age_; ++it, ++i);
    
    return (*it);
}

// age_ == 0 is current, otherwise from buffer
wxImage* PixalcoApp::GetImage(int row_, int age_) 
{
    // TODO: line should depend on line in which current algorithm runs
    int line = 0;

    if ( age_ < 0 )
        return NULL;

    if ( age_ == 0 )
    {
        ImageScope * scope = mMainFrame->GetImageScope(line, row_);
        if ( !scope )
            return NULL;
        return scope->GetImage();
    }
    else
    {
        ImageObject * obj = GetImageObjFromBuffer(line, row_, age_-1);
        if ( !obj )
            return NULL;

        return obj->GetImage();
    }
    
    return NULL;
}

void PixalcoApp::AddImageObjToMovieBuffer(ImageObject * imgObj_)
{
    if ( !imgObj_ )
        return;

//    cerr << imgObj_->GetDisplayName();

    imgObj_->AddRef();
    mMovieBuffer.push_front(imgObj_);
    if ( mMovieBuffer.size() > mMovieBufferSize )
    {
        mMovieBuffer.back()->Release();
        mMovieBuffer.pop_back();
    }     
}

wxImage* PixalcoApp::GetImageFromMovieBuffer(int age_)
{
    if ( mMovieBuffer.size() <= age_ )
        return NULL;

    list<ImageObject*>::iterator it=mMovieBuffer.begin();
    for ( int i=0; i < age_; ++it, ++i);
    
    return (*it)->GetImage();
}

void PixalcoApp::ClearBuffer(int line_, int row_)
{
    if ( row_ < 0 || row_ >= MAX_ROWS || line_ < 0 || line_ > 0 )
        return;
        
    while( mImageObjBuffer[row_]->size() > 0 )
    {
        mImageObjBuffer[row_]->back()->Release();
        mImageObjBuffer[row_]->pop_back();
    }
}

void PixalcoApp::ClearMovieBuffer()
{
    while ( mMovieBuffer.size() > 0 )
    {
        mMovieBuffer.front()->Release();
        mMovieBuffer.pop_front();
    }
}

void PixalcoApp::SetImageBufferSize(int size_)
{
    for ( int i=0; i<MAX_ROWS; ++i )
    {
        while( mImageObjBuffer[i]->size() > size_ )
        {
            mImageObjBuffer[i]->front()->Release();
            mImageObjBuffer[i]->pop_front();
        }
    }
    mImageBufferSize = size_;
}

void PixalcoApp::SetMovieBufferSize(int size_)
{
    while ( mMovieBuffer.size() > size_ )
    {
        mMovieBuffer.back()->Release();
        mMovieBuffer.pop_back();
    }
    mMovieBufferSize = size_;        
}
